From e0f307736ebd50d481c5b4ed781e78396977d3b3 Mon Sep 17 00:00:00 2001
From: Sungju Kwon <sungju.kwon@gmail.com>
Date: Tue, 25 Jun 2019 16:43:54 +1000
Subject: [PATCH] Allows to change the error output direction

Currently, the error() is always printing the output to the console
through 'stdout'. This does not follow redirection which is good when
you want to know error while redirecting commands output to a file.
However, there are situations that you want to hide error messages or
redirect it into somewhere else.

Using 'set stderr' command, it can be changed to three different
destination - fixed 'stdout', following redirection (fp), or a custom
file path.

        crash> set stderr
        stderr: stdout
        crash> sym 0x523 > /dev/null
        sym: invalid address: 0x523
        crash> set stderr fp
        stderr: fp
        crash> sym 0x523 > /dev/null
        crash> set stderr /tmp/err.log
        stderr: /tmp/err.log
        crash> sym 0x523 > /dev/null
        crash> set stderr stdout
        stderr: stdout
        crash> sym 0x523 > /dev/null
        sym: invalid address: 0x523
---
 defs.h  |  2 ++
 help.c  |  5 +++++
 main.c  |  2 ++
 tools.c | 70 +++++++++++++++++++++++++++++++++++++++++++++++++++++------------
 4 files changed, 66 insertions(+), 13 deletions(-)

diff --git a/defs.h b/defs.h
index ccffe58..cf46bfa 100644
--- a/defs.h
+++ b/defs.h
@@ -553,6 +553,8 @@ struct program_context {
 	ulong scope;			/* optional text context address */
 	ulong nr_hash_queues;		/* hash queue head count */
 	char *(*read_vmcoreinfo)(const char *);
+        FILE *stderr;                   /* error() message direction */
+        char *stderr_path;              /* stderr path information */
 };
 
 #define READMEM  pc->readmem
diff --git a/help.c b/help.c
index 581e616..ddc8e86 100644
--- a/help.c
+++ b/help.c
@@ -1093,6 +1093,10 @@ char *help_set[] = {
 "         redzone  on | off     if on, CONFIG_SLUB object addresses displayed by",
 "                               the kmem command will point to the SLAB_RED_ZONE",
 "                               padding inserted at the beginning of the object.", 
+" stderr  stdout | fp | <path>  set the direction of error put. 'stdout' always",
+"                               print on console. 'fp' follows the redirection",
+"                               or pipe command. '<path>' can be any file path",
+"                               in the filesystem which can save the output",
 " ",
 "  Internal variables may be set in four manners:\n",
 "    1. entering the set command in $HOME/.%src.",
@@ -1144,6 +1148,7 @@ char *help_set[] = {
 "             scope: (not set)",
 "           offline: show",
 "           redzone: on",
+"            stderr: stdout",
 " ",
 "  Show the current context:\n",
 "    %s> set",
diff --git a/main.c b/main.c
index 83ccd31..f80432a 100644
--- a/main.c
+++ b/main.c
@@ -1085,6 +1085,7 @@ setup_environment(int argc, char **argv)
 	 *  to pipes or output files.
 	 */
 	fp = stdout;
+        set_stderr("stdout");
 
 	/*
 	 *  Start populating the program_context structure.  It's used so
@@ -1725,6 +1726,7 @@ dump_program_context(void)
 		pc->scope ? "" : "(not set)");
 	fprintf(fp, "   nr_hash_queues: %ld\n", pc->nr_hash_queues);
 	fprintf(fp, "  read_vmcoreinfo: %lx\n", (ulong)pc->read_vmcoreinfo);
+        fprintf(fp, "           stderr: %s\n", pc->stderr_path);
 }
 
 char *
diff --git a/tools.c b/tools.c
index 2d95c3a..7d1e843 100644
--- a/tools.c
+++ b/tools.c
@@ -38,6 +38,43 @@ static void print_value(struct req_entry *, unsigned int, ulong, unsigned int);
 static struct req_entry *fill_member_offsets(char *);
 static void dump_struct_members_fast(struct req_entry *, int, ulong);
 
+FILE *
+set_stderr(char *target)
+{
+        FILE *tmp_fp = NULL;
+        char *tmp_str = NULL;
+
+        if (pc->stderr_path != NULL && !strcmp(target, pc->stderr_path))
+                return pc->stderr;
+
+        tmp_str = malloc(strlen(target) + 1);
+        if (tmp_str == NULL)
+                return NULL;
+        strcpy(tmp_str, target);
+
+        if (!strcmp(target, "stdout")) {
+                tmp_fp = stdout;
+        } else if (!strcmp(target, "fp")) {
+                tmp_fp = fp;
+        } else {
+                tmp_fp = fopen(target, "a");
+                if (tmp_fp == NULL) {
+                        error(INFO, "invalid path: %s\n", target);
+                        return NULL;
+                }
+        }
+
+        if (pc->stderr != NULL && pc->stderr != stdout && pc->stderr != fp)
+                fclose(pc->stderr);
+
+        if (pc->stderr_path)
+                free(pc->stderr_path);
+        pc->stderr = tmp_fp;
+        pc->stderr_path = tmp_str;
+        return pc->stderr;
+}
+
+
 /*
  *  General purpose error reporting routine.  Type INFO prints the message
  *  and returns.  Type FATAL aborts the command in progress, and longjmps
@@ -58,6 +95,9 @@ __error(int type, char *fmt, ...)
         void *retaddr[NUMBER_STACKFRAMES] = { 0 };
 	va_list ap;
 
+        if (!strcmp(pc->stderr_path, "fp"))
+                pc->stderr = fp;
+
 	if (CRASHDEBUG(1) || (pc->flags & DROP_CORE)) {
 		SAVE_RETURN_ADDRESS(retaddr);
 		console("error() trace: %lx => %lx => %lx => %lx\n",
@@ -69,7 +109,7 @@ __error(int type, char *fmt, ...)
         va_end(ap);
 
 	if (!fmt && FATAL_ERROR(type)) {
-		fprintf(stdout, "\n");
+		fprintf(pc->stderr, "\n");
 		clean_exit(1);
 	}
 
@@ -95,24 +135,14 @@ __error(int type, char *fmt, ...)
 			buf);
 		fflush(pc->stdpipe);
 	} else { 
-		fprintf(stdout, "%s%s%s %s%s", 
+		fprintf(pc->stderr, "%s%s%s %s%s", 
 			new_line || end_of_line ? "\n" : "",
 			type == WARNING ? "WARNING" : 
 			type == NOTE ? "NOTE" : 
 			type == CONT ? spacebuf : pc->curcmd,
 			type == CONT ? " " : ":",
 			buf, end_of_line ? "\n" : "");
-		fflush(stdout);
-	}
-
-        if ((fp != stdout) && (fp != pc->stdpipe) && (fp != pc->tmpfile)) {
-                fprintf(fp, "%s%s%s %s", new_line ? "\n" : "",
-			type == WARNING ? "WARNING" : 
-			type == NOTE ? "NOTE" : 
-			type == CONT ? spacebuf : pc->curcmd, 
-			type == CONT ? " " : ":",
-			buf);
-		fflush(fp);
+		fflush(pc->stderr);
 	}
 
 	if ((pc->flags & DROP_CORE) && (type != NOTE)) {
@@ -2483,6 +2513,19 @@ cmd_set(void)
 			}
 			return;
 
+                } else if (STREQ(args[optind], "stderr")) {
+                        if (args[optind+1]) {
+                                optind++;
+                                if (set_stderr(args[optind]) == NULL)
+                                        return;
+                        }
+
+                        if (runtime) {
+                                fprintf(fp, "stderr: %s\n",
+                                        pc->stderr_path);
+                        }
+                        return;
+
 		} else if (XEN_HYPER_MODE()) {
 			error(FATAL, "invalid argument for the Xen hypervisor\n");
 		} else if (pc->flags & MINIMAL_MODE) {
@@ -2590,6 +2633,7 @@ show_options(void)
 		fprintf(fp, "(not set)\n");
 	fprintf(fp, "       offline: %s\n", pc->flags2 & OFFLINE_HIDE ? "hide" : "show");
 	fprintf(fp, "       redzone: %s\n", pc->flags2 & REDZONE ? "on" : "off");
+	fprintf(fp, "        stderr: %s\n", pc->stderr_path);
 }
 
 
-- 
1.8.3.1

